package org.vaadin.peter.buttongroup;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import com.vaadin.ui.AbstractComponentContainer;
import com.vaadin.ui.Button;
import com.vaadin.ui.Component;

public class ButtonGroup extends AbstractComponentContainer {
    private static final long serialVersionUID = 5302911140929856951L;

    private final List<Button> buttons;

    public ButtonGroup() {
        buttons = new LinkedList<Button>();
    }

    /**
     * Adds given button to this button group. Group will be filled from left to
     * right.
     * 
     * @param button
     * @return given button
     * @throws IllegalArgumentException
     *             if given button already has a parent component
     */
    public Button addButton(Button button) {
        return addButton(button, buttons.size());
    }

    /**
     * Replaces given oldButton with given newButton. newButton will be
     * positioned to the same place from where old one was taken away.
     * 
     * @param oldButton
     * @param newButton
     * @throws IllegalArgumentException
     *             if given old button does not exist in this group
     * @throws IllegalArgumentException
     *             if given newButton already has a parent component
     */
    public void replaceButton(Button oldButton, Button newButton) {
        if (!buttons.contains(oldButton)) {
            throw new IllegalArgumentException("Given old button does not exist in this group");
        }

        int oldPosition = buttons.indexOf(oldButton);

        if (oldPosition == -1) {
            return;
        }

        removeComponent(oldButton);

        addButton(newButton, oldPosition);
    }

    /**
     * Removes given button from this group.
     * 
     * @param button
     */
    public void removeButton(Button button) {
        if (!buttons.contains(button)) {
            return;
        }

        buttons.remove(button);
        super.removeComponent(button);

        markAsDirty();
    }

    /**
     * Removes all buttons from this group
     */
    public void removeAllButtons() {
        super.removeAllComponents();
        buttons.clear();
    }

    private Button addButton(Button button, int position) {
        if (button.getParent() != null) {
            throw new IllegalArgumentException("Given button already has a parent component");
        }

        buttons.add(position, button);
        super.addComponent(button);

        markAsDirty();

        return button;
    }

    @Override
    public void addComponent(Component c) {
        throw new UnsupportedOperationException("Use addButton method instead");
    }

    @Override
    public void removeComponent(Component c) {
        throw new UnsupportedOperationException("Use removeButton method instead");
    }

    @Override
    public void removeAllComponents() {
        throw new UnsupportedOperationException("Use removeAllButtons instead");
    }

    @Override
    public Iterator<Component> iterator() {
        return new ArrayList<Component>(buttons).iterator();
    }

    @Override
    public void replaceComponent(Component oldComponent, Component newComponent) {
        throw new UnsupportedOperationException("Use replaceButton method instead");
    }

    @Override
    public int getComponentCount() {
        return buttons.size();
    }

}
